package com.cecf.observe;

import org.apache.flink.api.common.functions.RichMapFunction;
import org.apache.flink.configuration.Configuration;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.sql.*;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.locks.ReentrantReadWriteLock;


public class FillAppNameMapFunc extends RichMapFunction<SpringBootSpan, SpringBootSpan> {
    private Thread dictUpdateThread;

    private Connection mysqlConn;

    private final String mysqlDSN;
    private final String mysqlPwd;
    private final String mysqlUserName;

    public FillAppNameMapFunc(String dsn, String u, String pwd) {
        mysqlDSN = dsn;
        mysqlUserName = u;
        mysqlPwd = pwd;
    }

    public Connection connectMySQL() throws SQLException {

        try {
            return DriverManager.getConnection(mysqlDSN, mysqlUserName, mysqlPwd);
        } catch (SQLException e) {
            logger.error("连接mysql数据库失败:{}:{}@{}", mysqlUserName, mysqlPwd, mysqlDSN, e);
            throw e;
        }
    }

    private final ReentrantReadWriteLock mapLock = new ReentrantReadWriteLock();

    private Map<String, String> service2AppName = new HashMap<>();
    private boolean continueUpdateDict = true;
    public static final Logger logger = LoggerFactory.getLogger(FillAppNameMapFunc.class);

    @Override
    public void open(Configuration parameters) throws Exception {
        mysqlConn = connectMySQL();

        updateDictionary();
        dictUpdateThread = new Thread(() -> {
            while (continueUpdateDict) {
                try {
                    TimeUnit.SECONDS.sleep(600);
                } catch (InterruptedException e) {
                    logger.warn("sleep interrupted:", e);
                    continue;
                }
                updateDictionary();
            }
            logger.info("serviceName2AppName dict update thread quitting");
        });
        dictUpdateThread.start();
    }

    @Override
    public void close() throws Exception {
        continueUpdateDict = false;
        logger.info("stop begin, wait update thread");
        dictUpdateThread.join();
        logger.info("stop done");
    }

    @Override
    public SpringBootSpan map(SpringBootSpan value) {
        if ((value.AppName == null || value.AppName.isEmpty() || value.AppName.equals("UNSET"))) {
            mapLock.readLock().lock();
            value.AppName = service2AppName.getOrDefault(value.ServiceName, "UNSET");
            mapLock.readLock().unlock();
        }
        return value;
    }

    private boolean areHashMapsEqual(Map<String, String> map1, Map<String, String> map2) {
        if (map1 == null) {
            return map2 == null;
        }
        if (map2 == null) {
            return false;
        }
        if (map1.size() != map2.size()) {
            return false;
        }

        for (String key : map1.keySet()) {
            if (!map2.containsKey(key)) {
                return false;
            }

            String value1 = map1.get(key);
            String value2 = map2.get(key);

            if (!value1.equals(value2)) {
                return false;
            }
        }

        return true;
    }

    private void updateDictionary() {
        Map<String, String> newService2AppName = new HashMap<>();
        String query = "SELECT service_name, app_name FROM observe.data_service_appname where delete_at is null";
        try {
            PreparedStatement preparedStatement = mysqlConn.prepareStatement(query);
            ResultSet resultSet = preparedStatement.executeQuery();
            while (resultSet.next()) {
                String serviceName = resultSet.getString("service_name");
                String appName = resultSet.getString("app_name");
                newService2AppName.put(serviceName, appName);
            }
        } catch (SQLException e) {
            logger.error("查询service到appName的映射关系出错:", e);
            return;
        }
        if (areHashMapsEqual(newService2AppName, service2AppName)) {
            logger.info("serviceName2AppName表未改变，跳过更新");
            return;
        }

        mapLock.writeLock().lock();
        service2AppName = newService2AppName;
        logger.info("serviceName2AppName dict updated, size: {}", service2AppName.size());
        mapLock.writeLock().unlock();
    }

}