package com.cecf.observe;

import org.apache.flink.api.common.eventtime.WatermarkStrategy;
import org.apache.flink.api.common.serialization.SimpleStringSchema;
import org.apache.flink.api.java.utils.ParameterTool;
import org.apache.flink.configuration.Configuration;
import org.apache.flink.configuration.RestartStrategyOptions;
import org.apache.flink.connector.kafka.source.KafkaSource;
import org.apache.flink.connector.kafka.source.enumerator.initializer.OffsetsInitializer;
import org.apache.flink.streaming.api.environment.StreamExecutionEnvironment;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import ru.ivi.opensource.flinkclickhousesink.ClickHouseSink;
import ru.ivi.opensource.flinkclickhousesink.model.ClickHouseClusterSettings;
import ru.ivi.opensource.flinkclickhousesink.model.ClickHouseSinkConst;

import java.time.Duration;
import java.util.HashMap;
import java.util.Map;
import java.util.Properties;

public class TraceStream {

    private static String getEnv(String key, String de) {
        String value = System.getenv(key);
        if (value == null || value.isEmpty()) {
            return de;
        }
        return value;
    }

    public static final Logger logger = LoggerFactory.getLogger(TraceStream.class);

    static KafkaSource<String> buildKafkaSource() {
        String brokers = getEnv("KAFKA_BROKERS", "kafka-headless.observe.svc.cluster.local:9092");
        return KafkaSource.<String>builder()
                .setBootstrapServers(brokers)
                .setTopics(getEnv("KAFKA_TOPIC", "otel"))
                .setGroupId(getEnv("KAFKA_GROUP", "ob-agent-stream"))
                .setStartingOffsets(OffsetsInitializer.latest())
                .setValueOnlyDeserializer(new SimpleStringSchema())
                .build();
    }

    public static void main(String[] args) throws Exception {
        Configuration envConfig = new Configuration();
        envConfig.set(RestartStrategyOptions.RESTART_STRATEGY, "fixed-delay");
        envConfig.set(RestartStrategyOptions.RESTART_STRATEGY_FIXED_DELAY_ATTEMPTS, Integer.MAX_VALUE); // number of restart attempts
        envConfig.set(RestartStrategyOptions.RESTART_STRATEGY_FIXED_DELAY_DELAY, Duration.ofSeconds(10)); // delay
        StreamExecutionEnvironment env = StreamExecutionEnvironment.getExecutionEnvironment(envConfig);
        Map<String, String> globalParameters = new HashMap<>();

        String chHosts = getEnv("CH_HOST", "http://clickhouse.observe.svc.cluster.local:8123");
        globalParameters.put(ClickHouseClusterSettings.CLICKHOUSE_HOSTS, chHosts);
        globalParameters.put(ClickHouseClusterSettings.CLICKHOUSE_USER, getEnv("CH_USER", "default"));
        globalParameters.put(ClickHouseClusterSettings.CLICKHOUSE_PASSWORD, getEnv("CH_PWD", "cecf@cestong.com"));

        globalParameters.put(ClickHouseSinkConst.TIMEOUT_SEC, getEnv("CH_TIMEOUT", "10"));
        globalParameters.put(ClickHouseSinkConst.FAILED_RECORDS_PATH, "./failed_record");
        globalParameters.put(ClickHouseSinkConst.NUM_WRITERS, getEnv("CH_WRITER_NUM", "1"));
        globalParameters.put(ClickHouseSinkConst.NUM_RETRIES, getEnv("CH_RETRY_NUM", "3"));
        globalParameters.put(ClickHouseSinkConst.QUEUE_MAX_CAPACITY, getEnv("CH_WRITE_QUEUE_SIZE", "10"));
        globalParameters.put(ClickHouseSinkConst.IGNORING_CLICKHOUSE_SENDING_EXCEPTION_ENABLED, "true");

        String excludeService = getEnv("EXCLUDE_SERVICE", "");
        String[] excludeServiceList = excludeService.split(",");

        ParameterTool parameters = ParameterTool.fromMap(globalParameters);
        env.getConfig().setGlobalJobParameters(parameters);
        KafkaSource<String> source = buildKafkaSource();

        Properties props = new Properties();
        props.put(ClickHouseSinkConst.TARGET_TABLE_NAME, getEnv("CH_TABLE_NAME", "otel.traces_local"));
        props.put(ClickHouseSinkConst.MAX_BUFFER_SIZE, getEnv("CH_WRITE_BUF_SIZE", "1000000000"));

        SpringBootSpanCSVConverter converter = new SpringBootSpanCSVConverter();


        String mysqlDSN = getEnv("MYSQL_DSN", "jdbc:mysql://m1.cestong.com.cn:31511/observe");
        String mysqlUserName = getEnv("MYSQL_USER", "root");
        String mysqlPwd = getEnv("MYSQL_PWD", "pg3mfWRtYonekZWB");

        env.fromSource(source, WatermarkStrategy.noWatermarks(), "KafkaSource")
                .map(new SpringBootSpanExtractMapFunc())
                .name("InfoExtract")
                .flatMap(new SpringBootFlatMapFunc())
                .name("ResourceSpansFlatToSpan")
                .filter(new ServiceNameFilter(excludeServiceList))
                .name("FilterByServiceName")
                .map(new FillAppNameMapFunc(mysqlDSN, mysqlUserName, mysqlPwd))
                .name("AttachAppName")
                .addSink(new ClickHouseSink<>(props, converter))
                .name("ClickhouseSink");
        env.fromSource()

        logger.info("配置完成，运行flink");
        env.execute(getEnv("FLINK_JOB_NAME", "Ob stream operator"));
    }


}
